# Copyright (C) 2011 by Intevation GmbH
# Authors:
# Sascha L. Teichmann <sascha.teichmann@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details

"""RFC 2616 Sec. 3.6 compatible chunked transfer encoding"""

class ChunkedTransferEncodingWriter(object):
    """A class to write byte data in chunks in a RFC 2616 Sec. 3.6 compatible
    manner to an output.

    The chunk size and the method used to write out the data are
    to be given at construction time.
    """

    def __init__(self, write, block_size=4096):
        """Initialize the writer with a write out method and a chunk size."""

        self.write = write
        self.block_size = block_size
        self.blocks = []
        self.size = 0

    def append(self, block):
        """Add a byte array to the output. The data is chunked according
        to the specified size. If the length of the given data is smaller
        than the chunk size the data is hold back until subsequent calls
        to append() add enough bytes to flush it in the right chunk size.
        """

        write = self.write
        block_size = self.block_size
        blocks = self.blocks

        while True:
            size = len(block)
            if size == 0:
                break
            new_size = self.size + size
            if new_size < block_size:
                blocks.append(block)
                self.size = new_size
                break
            tail_size = new_size - block_size
            head_size = size - tail_size
            blocks.append(block[0:head_size])
            data = ''.join(blocks)
            del blocks[:]
            self.size = 0
            write("%x\r\n" % block_size)
            write(data)
            data = None
            write("\r\n")
            block = block[head_size:]

    def finish(self):
        """Flushes the remaining data (if any) to the output and
        signals the end of stream by sending a zero-sized chunk.
        """

        write = self.write
        if self.size:
            write("%x\r\n" % self.size)
            write(''.join(self.blocks))
            write("\r\n")
            del self.blocks[:]
            self.size = 0
        write("0\r\n\r\n")
