# Copyright (C) 2007, 2008, 2009, 2010 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Read the InteProxy config file."""

import sys
import urlparse
from urllib import splitport, splituser, splitpasswd
from ConfigParser import SafeConfigParser, NoOptionError, NoSectionError


# special object to indicate no default value
nodefault = object()

class Struct(object):

    def __init__(self, **kw):
        self.__dict__.update(**kw)


class Option(object):

    def __init__(self, name, default=nodefault, converter=lambda x: x,
                 attribute=None):
        self.name = name
        self.default = default
        self.converter = converter
        if not attribute:
            attribute = name
        self.attribute = attribute

    def has_default(self):
        return self.default is not nodefault


class InteProxyRule(object):

    def __init__(self, scheme, host, path, query="", fragment="",
                 port=None, user=None, password=None):
        self.scheme = scheme
        self.host = host
        self.path = path
        self.query = query
        self.fragment = fragment
        self.port = port
        self.user = user
        self.password = password

    def __repr__(self):
        return ("InteProxyRule(%r, %r, %r, query=%r, fragment=%r, port=%r,"
                " user=%r, password=%r)"
                % (self.scheme, self.host, self.path, self.query, self.fragment,
                   self.port, self.user, self.password))

    def __eq__(self, other):
        return (self.scheme == other.scheme
                and self.host == other.host
                and self.path == other.path
                and self.query == other.query
                and self.fragment == other.fragment
                and self.port == other.port
                and self.user == other.user
                and self.password == other.password)

    def __lt__(self, other):
        for attr in ["scheme", "host", "path", "query", "fragment", "port",
                     "user", "password"]:
            this_value = getattr(self, attr)
            other_value = getattr(other, attr)
            if this_value == other_value:
                continue
            else:
                return this_value < other_value



def parse_inteproxy_url(url):
    real_scheme, rest = url.split(":", 1)
    scheme, netloc, path, query, fragment = urlparse.urlsplit("http:" + rest)
    userpass, hostport = splituser(netloc)
    if userpass is not None:
        user, password = splitpasswd(userpass)
    else:
        user = password = None
    host, port = splitport(hostport)
    if port is not None:
        port = int(port)
    return InteProxyRule(scheme=real_scheme, host=host, port=port,
                         path=path, query=query, fragment=fragment, user=user,
                         password=password)


def convert_bool(s):
    s = s.lower()
    if s in ["yes", "true", "1"]:
        return True
    if s in ["no", "false", "0"]:
        return False
    raise ValueError("Config option has wrong format: %r" % s)



inteproxy_desc = [Option("http_proxy", default=None),
                  Option("https_proxy", default=None),
                  Option("show_terms_dialog", converter=convert_bool,
                         default=1),
                  Option("cacerts", default=None),
                  Option("do_certificate_validation", converter=convert_bool,
                         default=0)]

proxy_desc = [Option("host"),
              Option("port", default=80, converter=int),
              Option("username", default=None),
              Option("password", default=None)]

host_desc = [Option("host"),
             Option("path"),
             Option("class", attribute="scheme")]


def read_config_section(parser, section, item_desc, defaults=None,
                        settings_class=Struct):
    if defaults is None:
        defaults = dict()
    options = dict()
    for item in item_desc:
        try:
            value = item.converter(parser.get(section, item.name,
                                              vars=defaults))
        except (NoOptionError, NoSectionError):
            if item.has_default():
                value = item.default
            else:
                print >>sys.stderr, "Missing option %r in section %r" \
                      % (item.name, section)
                sys.exit(1)

        options[item.attribute] = value

    return settings_class(**options)


rules_section_name = "inteproxy-rules"

def read_config(filename):
    """Reads the InteProxy configuration from the file given by filename.
    """
    parser = SafeConfigParser()
    parser.read([filename])

    sections = set(parser.sections())
    config = read_config_section(parser, "inteproxy", inteproxy_desc)
    sections.discard("inteproxy")

    for proxy_attr in ("http_proxy", "https_proxy"):
        section = getattr(config, proxy_attr)
        if section is not None:
            setattr(config, proxy_attr, read_config_section(parser, section,
                                                            proxy_desc))
            sections.discard(section)

    config.rules = []

    if rules_section_name in sections:
        sections.discard(rules_section_name)
        raw_urls = parser.get(rules_section_name, "urls")
        for url in raw_urls.splitlines():
            if url:
                config.rules.append(parse_inteproxy_url(url))

    # handle old host sections
    for section in sections:
        config.rules.append(read_config_section(parser, section, host_desc,
                                                settings_class=InteProxyRule))


    return config
