# Copyright (C) 2008, 2009, 2010 by Intevation GmbH
#
# Based on code from
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/456195 which
# is Copyright (C) 2006 by Alessandro Budai
#
# This software may be used and distributed according to the terms
# of the Python License 2.3 or newer, see http://www.python.org/license

"""
Functions to establish network connections of various kinds.
In particular, the functions allow vanilla TCP connections and SSL
connections and HTTP CONNECT based connections combined in a more or
less arbitrary ways.

In addition there's a httplib.HTTPConnection sub-class that can be
instantiated with an existing socket object so that one doesn't have to
override any methods to handle special situations such as connections
via an HTTPS proxy using the HTTP CONNECT method.
"""

import socket
import httplib
import urllib
import re

# the ssl modules was introduced in Python 2.6.  If available, we use
# that, otherwise the older ssl support from the socket module
try:
    import ssl
except ImportError:
    ssl = None


default_ports = {'http' : 80, 'https' : 443}


def parse_netloc(scheme, netloc):
    """Parses netloc and returns (hostname, port).

    If the netloc string doesn't contain a port, the default port of
    scheme (either 'http' or 'https') is used.  If the port cannot be
    determined, a ValueError is raise.
    """
    host, port = urllib.splitport(netloc)
    if port is None:
        port = default_ports.get(scheme)
        if port is None:
            raise ValueError("Cannot determine port for URL scheme %r"
                             % scheme)
    else:
        port = int(port)
    return host, port


def log_debug(template, *args):
    pass


def connect_tcp(host, port, log_debug=log_debug):
    """Establish a TCP connection to port port on host.
    The return value is the socket object for the connection
    """
    msg = "getaddrinfo returns an empty list"
    for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        try:
            sock = socket.socket(af, socktype, proto)
            log_debug("connect: (%s, %s)", host, port)
            sock.connect(sa)
        except socket.error, msg:
            log_debug("connect fail:", host, port)
            if sock:
                sock.close()
            sock = None
            continue
        break
    if not sock:
        raise socket.error, msg

    return sock


def connect_http_connect(sock, real_host, real_port, headers, debuglevel=0):
    """Establish a connection through a HTTP proxy using the CONNECT method.

    The function assumes sock is a socket-like object already connected
    to the proxy by whatever means is suitable for the caller.  The
    function sends the CONNECT request for the given real_host and
    real_port, and ready the response.  If the response indicates
    success, the function returns sock.  The headers parameter should be
    a list of additional sending any given additional HTTP request
    header fields.
    """
    sock.send("CONNECT %s:%d HTTP/1.0\r\n" % (real_host, real_port))
    for item in headers:
        sock.send("%s: %s\r\n" % item)
    sock.send("\r\n")
    response = httplib.HTTPResponse(sock, method="CONNECT",
                                    debuglevel=debuglevel)
    (version, code, message) = response._read_status()
    if code != 200:
        close()
        raise socket.error, ("Proxy connection failed: %d %s"
                             % (code, message.strip()))
    while True:
        line = response.fp.readline()
        if not line:
            raise socket.error, ("Proxy connection failed:"
                                 " unexpected end of response header")

        if line == '\r\n':
            break

    return sock


def extract_identities_from_certificate(certificate):
    """Extracts the peer's identity from subjectAltName or subject field that
    is contained in a peer certificate.

    Both fields contain a tuple that contains a sequence of relative
    distinguished names (RDN) stored in the certificates data structure (see
    http://docs.python.org/library/ssl.html for more details about this
    structure). If there are dNSNames defined in subjectAltName, these
    identities are returned and the search for commonNames is skiped. In some
    cases there are iPAddress stored in subjectAltname. For this reason, the
    return value of this function is a tuple of valid domains and ips."""
    domains = []
    ips = []
    subjectAltName = certificate.get('subjectAltName', ())
    for key, value in subjectAltName:
        if key == 'DNS':
            domains.append(value)
        if key == 'IP':
            # XXX Not sure if the name of the field storing IP adresses is
            # really iPAddress
            ips.append(value)
    if len(domains) > 0:
        # It is not necessary to continue searching for 'commonName' fields in
        # the certificate, because there is at least one dNSName defined that
        # MUST be used as the identity of the peer (see
        # http://www.ietf.org/rfc/rfc2818.txt for further details) 
        return domains, ips

    subject = certificate.get('subject', ())
    for rdn in subject:
        for key, value in rdn:
            if key == 'commonName' or key == 'dNSName':
                domains.append(value)
    return domains, ips


def create_url_regex(domain):
    """This function creates a regular expression that might be used to match
    certain domains.

    If domain contains wildcard characters '*', these characters are replaced by
    a regular expression that matches a sequence of valid hostname characters
    (letters, digits and hyphers). If domain is an invalid hostname, None is
    returned."""
    wildcard_char ="*"
    wildcard_regex = "[a-zA-Z0-9-]*"

    regex = ""
    first = True
    parts = domain.split(".")
    for part in parts:
        if not part or part[0] != "*" and not part[0].isalnum() or \
            part[-1] != "*" and not part[-1].isalnum():
            return None
        if not first:
            regex += "\\."
        first = False

        if wildcard_char in part:
            regex += part.replace(wildcard_char, wildcard_regex)
        else:
            regex += re.escape(part)
    return regex


def is_remote_domain_valid(peercert, hostname, log_debug=log_debug):
    """This function might be used to verify that hostname matches a valid
    domain in peercert."""
    hostname = hostname.lower()
    parts = hostname.split(".")
    domains, ips = extract_identities_from_certificate(peercert)
    for domain in domains:
        if len(domain.split(".")) != len(parts):
            continue

        regex = create_url_regex(domain.lower())
        if regex and re.match(regex, hostname):
            return True
    for ip in ips:
        if hostname == ip:
            return True
    log_debug("Server domain %r is not valid." % hostname)
    return False


def connect_ssl(sock, keyfile=None, certfile=None, log_debug=log_debug,
                ca_certs=None, cert_required=True, remote_url=None):
    """Establish an SSL connection on the socket-like object sock.
    If successful, the function returns a new socket like object that
    uses SSL to encrypt the data.
    """
    log_debug("initiate ssl on %r", sock)
    if ssl is not None:
        cert_reqs = ssl.CERT_REQUIRED
        if not cert_required:
            cert_reqs = ssl.CERT_NONE
        wrapped_sock = ssl.wrap_socket(sock, keyfile, certfile,
                                       ca_certs=ca_certs,
                                       cert_reqs=cert_reqs)
        if not cert_required:
            return wrapped_sock

        peercert = wrapped_sock.getpeercert()
        if not peercert or not is_remote_domain_valid(peercert, remote_url,
                                                      log_debug):
            raise ssl.SSLError("Server domain does not match any identity "
                               "specified in the certificate.")
    else:
        sslsock = socket.ssl(sock, keyfile, certfile)
        wrapped_sock = httplib.FakeSocket(sock, sslsock)
    log_debug("ssl initiated on %r", sock)
    return wrapped_sock


class SocketHTTPConnection(httplib.HTTPConnection):

    """Extends HTTPConnection to accept a predefined socket-like object.
    The base class tries to establish the connection itself, which is
    awkward given all the different ways a connection can be established
    in inteproxy, such as directly per tcp, or SSL over an HTTPs proxy.
    """

    def __init__(self, sock, host, port=None, *args, **kw):
        httplib.HTTPConnection.__init__(self, host, port, *args, **kw)
        self.sock = sock

    def connect(self):
        pass
