#! /usr/bin/python
# Copyright (C) 2006, 2007, 2008, 2009, 2010 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""The HTTP proxy at the core of InteProxy"""

import sys
import traceback
import time
import urlparse
import BaseHTTPServer
import socket
import base64

# the ssl modules was introduced in Python 2.6.  If available, we use
# that, otherwise the older ssl support from the socket module
try:
    import ssl
except ImportError:
    ssl = None

from inteproxy.threadpool import ThreadPool
from inteproxy.feesdialog import handle_fees_and_access_constraints
from inteproxy.certificatedialog import handle_certificate_validation_error
from inteproxy.httpserver import HTTPServer
from inteproxy.httpmessage import HTTPRequestMessage, HTTPResponseMessage
from inteproxy.httpconnection import connect_tcp, connect_http_connect, \
     connect_ssl, SocketHTTPConnection, parse_netloc
from inteproxy.chunkedwriter import ChunkedTransferEncodingWriter


# same as the BaseHTTPRequestHandler method, but as a standalone function:
def log_date_time_string():
    """Return the current time formatted for logging."""
    now = time.time()
    year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
    s = "%02d/%3s/%04d %02d:%02d:%02d" % (
        day, BaseHTTPServer.BaseHTTPRequestHandler.monthname[month],
        year, hh, mm, ss)
    return s

class InteProxyHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    # Whether debug output should be logged.  Supported values:
    #     0    no debug output
    #     1    normal debug level with abbreviated log of the data
    #          transferred
    #   >=2    Also logs all the data transferred by the proxy
    #
    debug_level = 0

    def do_GET(self):
        """Handle a GET request"""
        self.handle_proxy_request()

    def do_POST(self):
        """Handle a POST request"""
        self.handle_proxy_request()

    def handle_proxy_request(self):
        """Handle a request that needs to be 'proxied'.
        """
        self.log_debug("Got %s request for %s, version '%s'",
                    self.command, self.path, self.request_version)
        for header, value in self.headers.items():
            self.log_debug("header from client: %s:%r", header, value)

        #
        #  Read the client request
        #
        client_request = self.read_client_request()

        #
        # Make sure that it requests compressed data
        #
        self.ensure_encoding_header(client_request)

        #
        # Determine the transcoder to use
        #
        transcoder = self.server.transcoder_map.get_transcoder(self.command,
                                                               self.path)

        #
        # convert request according to the transcoder rules
        #
        transcoder.convert_request(client_request)
        self.log_debug("modified client_request:")
        client_request.debug_log_message(self.log_debug)

        #
        # Create and send new request to the real remote host.
        #
        remote_url = transcoder.get_url()
        self.log_debug("Converted url: %r", remote_url)
        response = None
        try:
            response = self.perform_http_request(remote_url, client_request)
        except socket.error, exc:
            # some low level error occurred (e.g. the connection to an
            # upstream proxy could not be established).  Send a Bad
            # Gateway error to the client
            self.send_error(502)

        if response is not None:
            if self.should_decompress_response:
                response.do_decompress()

            # check for fees and access constraints and run a dialog
            if self.server.show_terms_dialog:
                handle_fees_and_access_constraints(remote_url, response)
            self.handle_response(response)

        self.log_debug("request finished")

    def read_client_request(self):
        """Read the client request including the body, if any.
        The request is returned as a TranscoderRequest instance."""
        return HTTPRequestMessage(self.command, self.path, self.request_version,
                                  self.headers, self.rfile)

    def create_proxy_headers(self, proxy):
        headers = []
        if proxy is not None:
            if proxy.username and proxy.password:
                userpass = base64.b64encode("%s:%s" % (proxy.username,
                                                       proxy.password))
                headers.append(("Proxy-Authorization",
                                "Basic %s" % userpass))
        return headers


    def https_connection_wrapper(self, remote_address):
        """Open a HTTPS connection to remote_address and validate the server
        certificate

        If the configuration option do_certificate_validation is set to
        true and if python >= 2.6 the server certificate is validated.
        If the validation failed, a popup dialog with technical details
        of the failure is shown.
        The user has the choice to trust or reject the connection. Rejecting the
        connection will result in a HTTP-502. If python < 2.6 is installed or
        if the configuration option do_certificate_validation is not set to
        true, the certificate is not validated. 
        An ssl connection is used in both cases to encrypt the data."""
        if ssl is not None:
            try:
                return self.open_https_connection(remote_address,
                                                  self.server.validate_cert)
            except ssl.SSLError:
                self.log_debug("SSL certificate validation failed.")
                if handle_certificate_validation_error(remote_address[0]):
                    return self.open_https_connection(remote_address,
                                                      validate=False)
                return None
        else:
            return self.open_https_connection(remote_address, validate=False)


    def open_https_connection(self, remote_address, validate=True):
        """Open a HTTPS connection to remote_address

        This method handles proxies as well. Its return value is a socket like
        object using SSL to encrypt the data. If python >= 2.6 is installed, the
        server certificate is validated. If this validation failed None is
        returned, otherwise a ssl like object.  The parameter validate may be
        used to skip the certificate validation.If python < 2.6 is installed,
        the certificate is not validated, but a ssl connection is used to
        encrypt the data."""
        self.log_debug("Open HTTPS connection to %r" % remote_address[0])
        sock = None

        proxy = self.server.https_proxy
        if proxy:
            sock = connect_tcp(proxy.host, proxy.port,
                               log_debug=self.log_debug)
            sock = connect_http_connect(sock, remote_address[0],
                                        remote_address[1],
                                        self.create_proxy_headers(proxy),
                                        debuglevel=self.debug_level)
        else:
            sock = connect_tcp(remote_address[0], remote_address[1],
                               log_debug=self.log_debug)

        if ssl is not None and validate:
            self.log_debug("Start HTTPS connection and validate certificates.")
            cert_file = self.server.get_cacert_path()
            return connect_ssl(sock, log_debug=self.log_debug,
                               ca_certs=cert_file, remote_url=remote_address[0])
        elif ssl is not None and not validate:
            self.log_debug("Start HTTPS connection but do not validate certificates.")
            return connect_ssl(sock, log_debug=self.log_debug,
                               cert_required=False)
        else:
            self.log_debug("Certificate validation is not supported.")
            return connect_ssl(sock, log_debug=self.log_debug)


    def open_http_connection(self, remote_url, remote_address, request_uri,
                             extra_headers):
        """Open a HTTP connection to remote_address

        This method handles proxies as well. Its return value is a socket
        object."""
        self.log_debug("Open HTTP connection to %r" % remote_address[0])

        proxy = self.server.http_proxy
        if proxy:
            extra_headers.extend(self.create_proxy_headers(proxy))
            return connect_tcp(proxy.host, proxy.port,
                               log_debug=self.log_debug), remote_url
        else:
            return connect_tcp(remote_address[0], remote_address[1],
                               log_debug=self.log_debug), request_uri


    def perform_http_request(self, remote_url, client_request):
        """Open a HTTP connection to remote_url and send client_request

        This method handles both http and https URLs as well as proxies
        for both.  The return value is a HTTPResponseMessage object
        containing the remote server's or proxy's response.
        """
        scheme, netloc, path, query, fragment = urlparse.urlsplit(remote_url)
        remote_address = parse_netloc(scheme, netloc)

        # the URI used in the request.  Usually, it's like remote_url
        # but with schem and netloc removed.
        request_uri = urlparse.urlunsplit(("", "", path, query, fragment))

        extra_headers = [("Host", "%s:%d" % remote_address)]


        sock = None

        if scheme == "http":
            sock, request_uri = self.open_http_connection(remote_url,
                                                          remote_address,
                                                          request_uri,
                                                          extra_headers)
        elif scheme == "https":
            sock = self.https_connection_wrapper(remote_address)
            if not sock:
                self.send_error(502)
                return

        connection = SocketHTTPConnection(sock, *remote_address)
        if self.debug_level > 1:
            connection.set_debuglevel(self.debug_level)
        if self.request_version == "HTTP/1.0":
            connection._http_vsn = 10

        connection.putrequest(client_request.method, request_uri,
                              skip_accept_encoding=True, skip_host=True)

        for header, value in client_request.headers.items():
            if header.lower() == "host":
                # the host header is already in extra_headers.  The
                # value we received from the client may have been wrong
                # anyway as in the old inteproxy scheme it refers to the
                # host this proxy is running on.
                continue
            else:
                connection.putheader(header, value)
        for header, value in extra_headers:
            connection.putheader(header, value)
        connection.endheaders()

        if client_request.body is not None:
            self.log_debug("Sending body to server: %r", client_request.body)
            connection.send(client_request.body)

        self.log_debug("request sent")

        response = connection.getresponse()
        for version_int, version in [(11, "HTTP/1.1"),
                                     (10, "HTTP/1.0"),
                                     ( 9, "HTTP/0.9")]:
            if response.version == version_int:
                break
        else:
            version = "UNKNOWN"
            self.log_debug("Unknown http version from server: %s",
                           response.version)
        if version != self.request_version:
            self.log_debug("Response version %r does not match"
                           " request version %r",
                           version, self.request_version)
        response_message = HTTPResponseMessage(version, response.status,
                                               response.reason,
                                               response.msg, response)
        response_message.debug_log_message(self.log_debug)

        # If an upstream server responds with 407, it's a proxy
        # requiring authentication.  We already sent all credentials we
        # know, so we can't do anything about it here, but we can send
        # an error to the inteproxy client.  We convert it to a 502
        # error, so that the client does not mistakenly think inteproxy
        # itself requires authentication
        if response_message.status == 407:
            self.log_debug("upstream sent %d: %r; converting to 502",
                           response_message.status, response_message.reason)
            response_message.status = 502
            response_message.reason = "Upstream proxy requires authentication"

        for header in ["Proxy-Authenticate"]:
            del response_message.headers[header]

        return response_message

    def ensure_encoding_header(self, client_request):
        """
        Request compression even if the client did not.
        This will modify the headers of client_request.
        """
        if not client_request.headers.get("Accept-Encoding"):
            client_request.headers["Accept-Encoding"] = "gzip, deflate"
            self.should_decompress_response = True
        elif ( not "gzip" in client_request.headers["Accept-Encoding"] and
               not "deflate" in client_request.headers["Accept-Encoding"] ):
            client_request.headers["Accept-Encoding"] = \
                ", ".join([client_request.headers["Accept-Encoding"],
                    "gzip", "deflate"])
            self.should_decompress_response = True
        else:
            self.should_decompress_response = False

    def send_headers(self, response):
        """Write the HTTP headers to the output stream."""
        for header, value in response.headers.items():
            self.log_debug("header to client: %s:%r", header, value)
            self.send_header(header, value)
        self.end_headers()

    def handle_response(self, response):
        # The HTTP version in the reply generated by send_response is
        # taken from self.protocol_version.  We simply set it to
        # whatever version was returned by the remote host.  That's what
        # the client would have received anyway, if it hadn't connected
        # via the InteProxy.
        self.protocol_version = response.version
        self.send_response(response.status, response.reason)

        do_rewrite = self.have_to_rewrite()
        do_chunked = response.headers.get("Transfer-encoding") == "chunked"

        if do_rewrite or self.should_decompress_response:
            response.do_decompress()

        if do_chunked and do_rewrite:
            self.send_headers(response)
            self.transfer_data_rewrite_chunked(response)
        else:
            if do_rewrite:
                self.rewrite_urls(response, do_rewrite)
            self.send_headers(response)
            self.transfer_data(response.read, self.wfile.write,
                               chunked = do_chunked)

    def transfer_data_rewrite_chunked(self, response):
        """Transfers the incoming data of the origin server in chunks
        and do url rewriting of the content at the same time.
        """

        transcoder_map = self.server.transcoder_map
        prefix = self.server.get_inteproxy_url()
        rewrite = transcoder_map.url_rewriter(prefix, self.log_debug)
        self.transfer_chunked_rewrite(rewrite, response.read, self.wfile.write)

    def wrap_read_write_debug(self, read, write):
        # wrap the read/write functions if debug logging is active so
        # that the data read from the server and written to the client
        # is logged.
        if self.debug_level > 0:
            orig_read = read
            orig_write = write

            def limit_length(data):
                """Returns a shortened version of the string data"""
                if self.debug_level < 2 and len(data) > 13:
                    data = "%s ... %x bytes ... %s" % (data[:5],
                                                       len(data) - 10,
                                                       data[-5:])
                return data

            def read(length):
                data = orig_read(length)
                self.log_debug("from server: %r", limit_length(data))
                return data
            def write(data):
                self.log_debug("to client: %r", limit_length(data))
                orig_write(data)

        return read, write

    def transfer_chunked_rewrite(self, rewrite, read, write,
                                 separator='>', length=4096):
        """Transfers data from read() to write() in chunks. The
        data is splitted by a given separator.
        """

        read, write = self.wrap_read_write_debug(read, write)
        writer = ChunkedTransferEncodingWriter(write, length)
        write = writer.append

        data = []
        append = data.append

        while True:
            chunk = read(length)
            if not chunk:
                break

            pos = 0
            while True:
                idx = chunk.rfind(separator, pos)

                if idx > 0:
                    rest = chunk[pos:idx]
                    append(rest)
                    rewritten = rewrite(''.join(data))
                    del data[:]
                    append(separator)
                    write(rewritten)
                    rewritten = None
                    pos = idx+1
                else:
                    if pos:
                        append(chunk[pos:])
                    else:
                        append(chunk)
                    break

        rewritten = rewrite(''.join(data))
        write(rewritten)
        rewritten = None
        writer.finish()

    def transfer_data(self, read, write, length=None, chunked=False):
        """Transfer data from one 'file' to another in chunks

        The files are given by their read and write methods so it
        doesn't have to be a file.  The read parameter must be callable
        with an integer argument indicating the maximum number of bytes
        to read and the write parameter must be callable with a string.
        If the parameter chunked is true, the method uses the 'chunked'
        transfer encoding when writing the data.
        """

        read, write = self.wrap_read_write_debug(read, write)

        # Now transfer the data in blocks of max_chunk_size
        max_chunk_size = 4096
        while 1:
            if length is not None:
                chunk_size = min(length, max_chunk_size)
            else:
                chunk_size = max_chunk_size
            chunk = read(chunk_size)
            if not chunk:
                break
            if length is not None:
                length -= len(chunk)
            if chunked:
                write("%x\r\n" % len(chunk))
            write(chunk)
            if chunked:
                write("\r\n")
        if chunked:
            write("0\r\n\r\n")

    def have_to_rewrite(self):
        """Returns if url rewriting is necessary."""
        return self.server.rewrite_urls and not urlparse.urlsplit(self.path)[0]


    def rewrite_urls(self, response, force_rewrite=False):
        """Rewrites URLs in the response if enabled in the server

        This method rewrites URLs in the response if the request is a
        old-style InteProxy request -- one where InteProxy is not used
        as a normal HTTP proxy -- and rewriting has been activated in
        the server.  The actual rewriting is done by the server's
        transcoder_map.
        """
        if force_rewrite or self.have_to_rewrite():
            transcoder_map = self.server.transcoder_map
            prefix = self.server.get_inteproxy_url()
            response.body = transcoder_map.rewrite_urls(response.body, prefix,
                                                        self.log_debug)

    def log_exception(self, exc_info):
        """Log an exception

        This method produces a log message via self.log_message and then
        prints a the exception given by exc_info, which should be the
        return value of sys.exc_info, to stderr.  By default log_message
        also writes to stderr, so both the message and the traceback
        should go to the same place.
        """
        self.log_message("Exception for previous 500 code\n")
        traceback.print_exception(*(exc_info + (None, sys.stderr)))
        sys.stderr.flush()

    def log_debug(self, template, *args):
        if self.debug_level > 0:
            message = "DEBUG: " + template % args
            message = message.replace("%", "%%")
            self.log_message(message)


class MasterWorkerServer(HTTPServer):

    do_shutdown = False

    def __init__(self, server_address, RequestHandlerClass, num_workers,
                 http_proxy, https_proxy, transcoder_map,
                 rewrite_urls=False, show_terms_dialog=True, cacert_path=None,
                 validate_cert=False):
        HTTPServer.__init__(self, server_address, RequestHandlerClass)
        self.http_proxy = http_proxy
        self.https_proxy = https_proxy
        self.transcoder_map = transcoder_map
        self.rewrite_urls = rewrite_urls
        self.show_terms_dialog = show_terms_dialog
        self.cacert_path = cacert_path
        self.validate_cert = validate_cert
        self.thread_pool = ThreadPool(num_workers, lambda f: f())
        sys.stderr.write("[%s] starting %d worker threads\n" \
                         % (log_date_time_string(), num_workers))
        self.thread_pool.start()

    def get_inteproxy_url(self):
        """Returns the URL of this InteProxy"""
        return "http://localhost:%d/" % self.getsockname()[1]

    def get_cacert_path(self):
        """Returns the path to a local cacert file"""
        return self.cacert_path

    def process_request(self, request, client_address):
        """Put the request into the queue to be handled by the worker thread
        """
        def process_in_worker():
            try:
                self.finish_request(request, client_address)
                self.close_request(request)
            except:
                self.handle_error(request, client_address)
                self.close_request(request)
        sys.stderr.write("[%s] queue contains %d items\n" %
                         (log_date_time_string(),
                          self.thread_pool.queue.qsize()))
        self.thread_pool.put(process_in_worker)

    def handle_error(self, request, client_address):
        """Override to integrate the error reporting better with the other logs
        """
        sys.stderr.write("[%s] Exception while handling request from %r\n"
                         % (log_date_time_string(), client_address))
        traceback.print_exception(*(sys.exc_info() + (None, sys.stderr)))
        sys.stderr.flush()

