# Copyright (C) 2008, 2009, 2011 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Tests for the inteproxy.config module"""

import sys
import os
import operator
import unittest

from filesupport import FileTestMixin
from support import AttributeTestMixin

from inteproxy.config import read_config, InteProxyRule


class ReadConfigTest(unittest.TestCase, FileTestMixin, AttributeTestMixin):

    def setUp(self):
        self.config_file = self.create_temp_file("test.cfg",
                                                 self.config_contents)


class TestReadConfigEmptyInteproxySection(ReadConfigTest):

    config_contents = """\
[inteproxy]
"""

    def test_readconfig(self):
        self.check_attributes(read_config(self.config_file),
                              http_proxy=None,
                              https_proxy=None)


class TestReadConfigNoInteproxySection(ReadConfigTest):

    config_contents = """\
"""

    def test_readconfig(self):
        self.check_attributes(read_config(self.config_file),
                              http_proxy=None,
                              https_proxy=None)


class TestReadConfigHttpProxyNoPortNoAuth(ReadConfigTest):

    config_contents = """\
[inteproxy]
http_proxy=local_proxy

[local_proxy]
host=localhost
"""

    def test_readconfig(self):
        config = read_config(self.config_file)
        self.check_attributes(config, https_proxy=None)
        self.check_attributes(config.http_proxy,
                              host="localhost", port=80,
                              username=None, password=None)


class TestReadConfigHttpProxySpecialChars(ReadConfigTest):

    config_contents = """\
[inteproxy]
http_proxy=special_proxy

[special_proxy]
host=example.com
username=foo\\b\ta\040r
password=special%%\()
"""

    def test_readconfig(self):
        config = read_config(self.config_file)
        self.check_attributes(config, https_proxy=None)
        self.check_attributes(config.http_proxy,
                              host="example.com", port=80,
                              username="foo\\b\ta r", password="special%\\()")


class TestReadConfigHttpsProxyWithPortAndAuth(ReadConfigTest):

    config_contents = """\
[inteproxy]
https_proxy=https_proxy

[https_proxy]
host=localhost
port=8080
username=john
password=secret
"""

    def test_readconfig(self):
        config = read_config(self.config_file)
        self.check_attributes(config, http_proxy=None)
        self.check_attributes(config.https_proxy,
                              host="localhost", port=8080,
                              username="john", password="secret")

class TestReadConfigProxyAndOldHosts(ReadConfigTest):

    config_contents = """\
[inteproxy]
https_proxy=https_proxy

[https_proxy]
host=localhost
port=8080
username=john
password=secret

[inteproxy-demo.intevation.org]
host=inteproxy-demo.intevation.org
path=/cgi-bin/frida-wms
class=owsproxy

[wms.example.org]
host=wms.example.org
path=/wms
class=basicauth
"""

    def test_readconfig(self):
        config = read_config(self.config_file)
        self.check_attributes(config, http_proxy=None)
        self.check_attributes(config.https_proxy,
                              host="localhost", port=8080,
                              username="john", password="secret")
        self.assertEquals(sorted(config.rules),
                          [InteProxyRule("basicauth",
                                         "wms.example.org",
                                         "/wms"),
                           InteProxyRule("owsproxy",
                                         "inteproxy-demo.intevation.org",
                                         "/cgi-bin/frida-wms"),
                           ])


class TestReadConfigNewHosts(ReadConfigTest):

    config_contents = """\
[inteproxy-rules]
urls=owsproxy://inteproxy-demo.intevation.org/cgi-bin/frida-wms
     basicauth://wms.example.org/wms
# comments are allowed
     basicauth://bob:tomato@data.intevation.*/
"""

    def test_readconfig(self):
        config = read_config(self.config_file)
        self.check_attributes(config, http_proxy=None, https_proxy=None)
        self.assertEquals(sorted(config.rules),
                          [InteProxyRule("basicauth", "data.intevation.*", "/",
                                         user="bob", password="tomato"),
                           InteProxyRule("basicauth",
                                         "wms.example.org",
                                         "/wms"),
                           InteProxyRule("owsproxy",
                                         "inteproxy-demo.intevation.org",
                                         "/cgi-bin/frida-wms"),
                           ])
